// 先准备fields, 钻取的行列fields, 画布的高度与宽度,
import * as _ from "lodash";

export class EchartsGridOptions {
    fields: string[];
    columns: string[];
    rows: string[];
    dataset: any[];
    cellWidth: number;
    cellHeight: number;
    canvasWidth: number;
    canvasHeight: number;
    seriesOptions: any;
}

export function init(options: EchartsGridOptions) {
    const getDatum = (fields: string[], keys: { [key: number]: any }, start: number) => {
        const fieldIndex = options.fields.findIndex(field => field === fields[start]);
        start++;
        const datum: {
            keys: string[],
            map: Map<string, any>,
            level?: number,
            dataset: any[],
        } = {
            keys: [],
            map: new Map<string, any>(),
            level: 1,
            dataset: options.dataset
        };
        const dataset = options.dataset.filter((di) => {
            const d = di.d;
            let match: boolean = true;
            if (keys) {
                for (let key of _.keys(keys)) {
                    match = d[key] === keys[key];
                    if (!match)
                        break;
                }
            }
            if (match && fieldIndex > -1) {
                if (datum.keys.indexOf(d[fieldIndex]) < 0) {
                    datum.keys.push(d[fieldIndex]);
                }
            }
            return match;
        });
        for (let key of datum.keys) {
            let addKey: any = {};
            addKey[fieldIndex] = key;
            datum.map[key] = getDatum(fields, {...keys, ...addKey}, start);
        }
        if (datum.keys[0]) {
            datum.level += datum.map[datum.keys[0]].level;
        }
        if (fieldIndex === -1) {
            datum.dataset = dataset;
            return datum;
        }
        return datum;
    };

    //增加rowKey
    options.dataset = options.dataset.map((d, i) => ({
        d,
        i
    }));

    const columns = getDatum(options.columns.slice(0, options.columns.length - 1), null, 0);
    const rows = getDatum(options.rows.slice(0, options.rows.length - 1), null, 0);
    const rowsMeta: {
        height: number;
        dataset: any[];
    }[] = [];
    const columnsMeta: {
        width: number;
        dataset: any[];
    }[] = [];

    const columnsHeight = columns.level * options.cellHeight;
    const rowsWidth = rows.level * options.cellWidth;
    const graphics: any[] = [];
    const grids: any[] = [];
    const series: any[] = [];
    const xAxis: any[] = [];
    const yAxis: any[] = [];

    const xField = options.rows.length > 0 ? options.rows[options.rows.length - 1] : null;
    const yField = options.columns.length > 0 ? options.columns[options.columns.length - 1] : null;
    //利用echarts的graph画列头与行头
    const canvasWidth = options.canvasWidth;
    const canvasHeight = options.canvasHeight;
    const height = options.cellHeight;
    const width = options.cellWidth;

    function createCharts(xField: string, yField: string, xAxis: any[], yAxis: any[], graphics: any[], grids: any[], series: any[], offsetX: number, offsetY: number) {
        let startY = 0;
        let startX = 0;
        let gridIndex = 0;
        let xFieldIndex = options.fields.findIndex(f => f === xField);
        let yFieldIndex = options.fields.findIndex(f => f === yField);
        columnsMeta.forEach((c, columnIndex) => {
            startY = 0;
            rowsMeta.forEach((r, rowIndex) => {
                // 数据集取交集
                const dataset = c.dataset.filter(d => r.dataset.findIndex(rd => rd.i === d.i) >= 0).map(d => d.d);

                const gridOption: any = {
                    x: offsetX + startX,
                    y: offsetY + startY + 10,
                    width: c.width - 10,
                    height: r.height - 20,
                    borderWidth: 0
                };

                const showXAxis = rowIndex == rowsMeta.length - 1;
                if (showXAxis) {
                    gridOption.height += 10;
                }
                xAxis.push({
                    type: 'category',
                    gridIndex,
                    show: showXAxis,
                    data: dataset.map(d => d[xFieldIndex])
                });

                const showYAxis = columnIndex == 0;
                yAxis.push({type: 'value', gridIndex, axisLabel: {show: showYAxis}, axisTick: {show: showYAxis}});
                grids.push(gridOption);
                graphics.push({
                    type: 'rect',
                    left: offsetX + startX,
                    top: offsetY + startY,
                    z: 100,
                    shape: {
                        width: c.width,
                        height: r.height
                    },
                    style: {
                        fill: 'transparent',
                        stroke: "#000"
                    }
                });
                series.push({
                    xAxisIndex: gridIndex,
                    yAxisIndex: gridIndex,
                    data: dataset.map(d => d[yFieldIndex]),
                    ...options.seriesOptions
                });
                gridIndex++;
                startY += r.height;
            });
            startX += c.width;
        });
    }

    function createColumnHeader(graphics: any[], width: number, height: number, text: string, offsetX: number) {
        if (text) {
            graphics.push({
                type: 'group',
                left: offsetX,
                top: 0,
                width,
                height,
                children: [
                    {
                        type: 'rect',
                        left: 0,
                        top: 0,
                        shape: {width, height},
                        style: {fill: "transparent"}
                    },
                    {type: 'text', left: 'center', top: 'center', style: {fill: '#000', text}}
                ]
            });
        }
    }

    function createRowsHeader(fields: string[], graphics: any[], width: number, height: number, cellWidth: number, offsetY: number) {
        const cellHeight = height - offsetY;
        const childGraphic: any = {
            type: 'group', left: 2, top: offsetY + 2, width, cellHeight,
            children: []
        };
        fields.forEach((text, i) => {
            childGraphic.children.push({
                type: 'group',
                left: i * cellWidth,
                top: 0,
                width: cellWidth,
                height: cellHeight,
                children: [
                    {
                        type: 'rect',
                        left: 0,
                        top: 0,
                        shape: {width: cellWidth, height: cellHeight},
                        style: {fill: "transparent"}
                    },
                    {
                        type: 'line',
                        left: 0,
                        top: cellHeight,
                        shape: {x1: 0, x2: cellWidth},
                        style: {fill: "#000"}
                    },
                    {type: 'text', left: 'center', top: 'center', style: {fill: '#000', text}}
                ]
            });
        });
        childGraphic.children.push({
            type: 'group',
            left: fields.length * cellWidth,
            top: 0,
            width: cellWidth,
            height: cellHeight,
            children: [
                {
                    type: 'rect',
                    left: 0,
                    top: 0,
                    shape: {width: cellWidth, height: cellHeight},
                    style: {fill: "transparent"}
                },
                {
                    type: 'line',
                    left: 0,
                    top: cellHeight,
                    shape: {x1: 0, x2: cellWidth},
                    style: {fill: "#000"}
                }
            ]
        });
        graphics.push(childGraphic);
    }

    function createColumnDimensions(columns: { keys: string[], map: Map<string, any>, dataset: any[] }, graphics: any[], level: number, width: number, height: number, offsetX: number, offsetY: number) {
        const columnWidth = columns.keys.length == 0 ? width : width / columns.keys.length;
        const nextLevel = level + 1;
        columns.keys.forEach((key, i) => {
            graphics.push({
                type: 'group',
                left: offsetX + i * columnWidth,
                top: offsetY + level * height,
                width: columnWidth,
                height,
                children: [
                    {
                        type: 'rect',
                        left: 0,
                        top: 0,
                        shape: {width: columnWidth, height},
                        style: {fill: 'transparent'}
                    },
                    {
                        type: 'line',
                        left: columnWidth,
                        top: 0,
                        shape: {x1: 0, y1: 0, x2: 0, y2: height},
                        style: {stroke: '#000'}
                    },
                    {type: 'text', left: 'center', top: 'center', style: {fill: '#000', text: key}},
                ]
            });

            if (columns.map[key]) {
                createColumnDimensions(columns.map[key] as any, graphics, nextLevel, columnWidth, height, offsetX + i * columnWidth, offsetY);
            } else {
                //维度集
                columnsMeta.push({
                    width: columnWidth,
                    dataset: columns.dataset
                });
            }
        });

        // 不存在keys， 直接在右端画横线
        if (columns.keys.length <= 0) {
            graphics.push({
                type: 'group',
                left: offsetX,
                top: offsetY + nextLevel * height,
                width: columnWidth,
                height,
                children: [
                    {
                        type: 'line',
                        left: columnWidth,
                        top: 0,
                        shape: {x1: 0, y1: 0, x2: 0, y2: height},
                        style: {stroke: '#000'}
                    }
                ]
            });
            columnsMeta.push({
                width: columnWidth,
                dataset: columns.dataset
            });
        }
    }

    function createRowDimensions(columns: { keys: string[], map: Map<string, any>, dataset: any[] }, graphics: any[], level: number, height: number, width: number, offsetX: number, offsetY: number) {
        const columnHeight = columns.keys.length == 0 ? height : height / columns.keys.length;
        const nextLevel = level + 1;
        columns.keys.forEach((key, i) => {
            graphics.push({
                type: 'group', left: level * width, top: offsetY + i * columnHeight, width, height: columnHeight,
                children: [
                    {
                        type: 'rect',
                        left: '0',
                        top: 0,
                        shape: {width, height: columnHeight},
                        style: {fill: 'transparent'}
                    },
                    {
                        type: 'line',
                        left: '0',
                        top: columnHeight,
                        shape: {x1: 0, y1: 0, x2: width, y2: 0},
                        style: {stroke: '#000'}
                    },
                    {type: 'text', left: 'center', top: 'center', style: {fill: '#000', text: key}}
                ]
            });

            if (columns.map[key]) {
                createRowDimensions(columns.map[key] as any, graphics, nextLevel, columnHeight, width, offsetX, offsetY + i * columnHeight);
            } else {
                //维度集
                rowsMeta.push({
                    height: columnHeight,
                    dataset: columns.dataset
                });
            }
        });
        if (columns.keys.length <= 0) {
            graphics.push({
                type: 'group', left: level * width, top: offsetY, width: 0, height: columnHeight,
                children: [
                    {
                        type: 'rect',
                        left: 0,
                        top: 0,
                        shape: {width, height: columnHeight},
                        style: {fill: 'transparent'}
                    },
                    {
                        type: 'line',
                        left: width,
                        top: 0,
                        shape: {x1: 0, y1: 0, x2: 0, y2: columnHeight},
                        style: {stroke: '#000'}
                    },
                    {
                        type: 'line',
                        left: 0,
                        top: columnHeight,
                        shape: {x1: 0, y1: 0, x2: width, y2: 0},
                        style: {stroke: '#000'}
                    }
                ]
            });
            rowsMeta.push({
                height: columnHeight,
                dataset: columns.dataset
            });
        }
    }


    createColumnHeader(graphics, canvasWidth - rowsWidth, height, options.columns.splice(0, options.columns.length - 1).join(" / "), rowsWidth);
    createColumnDimensions(columns, graphics, 0, canvasWidth - rowsWidth, height, rowsWidth, height);

    createRowsHeader(options.rows.splice(0, options.rows.length - 1), graphics, rowsWidth, columnsHeight, width, columnsHeight - height);
    createRowDimensions(rows, graphics, 0, canvasHeight - columnsHeight - height, width, 0, columnsHeight);
    createCharts(xField, yField, xAxis, yAxis, graphics, grids, series, rowsWidth + 2, columnsHeight + 2);

    return {
        grid: grids,
        series,
        graphic: graphics,
        yAxis,
        xAxis
    }
}
